//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DumpRom.C
//
//		Uses the TI Lynx 1394 API to dump the configuration ROM of a specified
//		1394 peripheral.
//
//		Use PCI Lynx 2.1 or 2.2 kit availible on the TI web site.
//		Download the kit from:
//
//			http://www.ti.com/sc/data/msp/1394/evm/lynx21.exe
//			http://www.ti.com/sc/data/msp/1394/evm/lynx22.exe
//
//		You can find information the TI Lynx Adapter (TSBKPCI) from:
//
//			http://www.ti.com/sc/docs/msp/1394/evm/pci403.htm
//
//		Set up the directories as follows:
//
//		..\DumpRom		The source to this file
//		..\inc			copy 1394api.h and 1394type.h from PCI Lynx kit 2.X (e.g. c:\lynxsoft)
//		..\lib			copy cls1394.lib from PCI Lynx kit 2.1 (e.g. c:\lynxsoft)
//		..\bin			copy cls1394.dll from PCI Lynx kit 2.1 (e.g. c:\lynxsoft)
//							we also copy the dumprom.exe file there for ease of execution
//							and debugging.
//		..\DumpRom.dsw	main project file for VC 5.0
//
//		Developed using MSVC 5.0.
//
//	10/13/98 Bill Alexander
//
//		Created.
//
//	1/19/99 Bill Alexander
//	
//		Added Get_Device_Info() call to get the device number of the first or specified
//		device.
//
//	1/20/99 Bill Alexander
//
//		Added configuration ROM decode.
//
//	1/25/99 Bill Alexander
//
//		Made it all work with 2.1 or 2.2 LynxSoft.
//
//	1/31/99
//
//		Added more comments to get ready for Dr. Dobbs Publishing.
//
//
//	Created with tabs set to every 3 spaces (i.e. 4, 7, ...)
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define _APP_MAIN_MODULE_
#define _DUMPROM_C_

#define STRICT
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <windows.h>

//
// The 1394api.h header file is provided by Texas Instruments to be
// with their Windows 95 API software.  The software is available on
// TI's website: http://www.ti.com
//
#include "..\inc\1394api.h"

// My local include files
#include "dumprom.h"
#include "dbg.h"

#pragma pack()

// LoadLibrary definitions
HANDLE hDll;
typedef BOOL (CALLBACK* PCls1394InitializeT)(DWORD *);
PCls1394InitializeT lpfnCls1394Initialize;
typedef BOOL (CALLBACK* PCls1394TerminateT)(void);
PCls1394TerminateT lpfnCls1394Terminate;
typedef BOOL (CALLBACK* PCls1394CreateFileT)(ULONG VendorID_DeviceType, WORD DeviceEntry);
PCls1394CreateFileT lpfnCls1394CreateFile;
typedef BOOL (CALLBACK* PCls1394CloseHandleT)(cls1394HANDLE hHnd);
PCls1394CloseHandleT lpfnCls1394CloseHandle;
typedef BOOL (CALLBACK* PCls1394DeviceIoControlT)(
									cls1394HANDLE  hDevice,
									DWORD  dwIoControlCode,
									LPVOID  lpInBuffer,
									DWORD  nInBufferSize,
									LPVOID  lpOutBuffer,
									DWORD  nOutBufferSize,
									LPDWORD  lpBytesReturned,
									LPOVERLAPPED  lpOverlapped);
PCls1394DeviceIoControlT lpfnCls1394DeviceIoControl;
typedef BOOL (CALLBACK* PGetDeviceInfoT)(WORD DeviceEntry, DeviceInfo* sDevInfo);
PGetDeviceInfoT lpfnGetDeviceInfo;


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Main
//
//		Main entry point to the program.
//
//	Entry:
//
//		argc			number of arguments in argv[] array
//		argv[]		array of strings containing command line arguments
//
//	Exit:
//
//		always returns 0
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int main(int argc, char *argv[])
{
	BOOL ccode;

	DebugFlags = DB_ALL;

	MyPrintf(VL_MAJOR_STEP, "1394 Dump ROM Utility\n");

	// Init vars
	ghDev = 0;
	gDriverInitialized = FALSE;
	gCfgRomBuf = NULL;
	gCfgRomBufSize = 0;

	// Process the command line
	ccode = ProcessCommandLine (argc, argv);
	if (ccode != TRUE)
		goto CommandLineError;

	// Load library and fix up functions pointers
	ccode = LoadLibraryAndFixPointers ();
	if (ccode != TRUE)
		goto LibraryError;

	// Initialize 1394 driver and DLL and reset the bus.
	// NOTE:  pass NULL as parameter for 2.1 and 2.2 compatibility
	MyPrintf (VL_MAJOR_STEP, "Initializing 1394 drivers ... ");
	ccode = lpfnCls1394Initialize (NULL);
	if (ccode != TRUE)
		goto DriverInitError;
	MyPrintf (VL_MAJOR_STEP, "OK\n");
	gDriverInitialized = TRUE;

	// Open the specified 1394 device
	MyPrintf (VL_MAJOR_STEP, "Opening 1394 Device ... ");
	ccode = DeviceOpen (gDeviceMode, &gDeviceNumber, &gDeviceId, &ghDev);
	if (ccode != TRUE)
		goto DevOpenError;
	MyPrintf (VL_MAJOR_STEP, "OK\n");

	// Are we dumping a specific quadlet?
	if (gUseSpecifiedOffset == TRUE)
	{
		// Dump the specific quadlet in config ROM
		ccode = DumpSpecificQuadlet (ghDev, gSpecifiedOffset);
		if (ccode != TRUE)
			goto ReadError;
	}
	else
	{
		// Dump entire configuration ROM

		// Allocate config rom buffer
		MyPrintf (VL_MAJOR_STEP, "Allocating configuration ROM buffer resources ... ");
		ccode = AllocateConfigRomBuffer (ghDev, &gCfgRomBuf, &gCfgRomBufSize);
		if (ccode != TRUE)
			goto AllocError;
		MyPrintf (VL_MAJOR_STEP, "OK\n");

		// Read the config rom into the buffer
		MyPrintf (VL_MAJOR_STEP, "Reading configuration ROM ... ");
		ccode = ReadConfigRomBuffer (ghDev, gCfgRomBuf, gCfgRomBufSize);
		if (ccode != TRUE)
			goto ReadError;
		MyPrintf (VL_MAJOR_STEP, "OK\n");

		// Dump Config ROM
		ccode = DumpConfigRom (gCfgRomBuf, gCfgRomBufSize);
		if (ccode != TRUE)
			goto Exit;

		// Decode Config ROM
		ccode = DecodeConfigRom (gCfgRomBuf, gCfgRomBufSize);
		if (ccode != TRUE)
			goto Exit;
	}

Exit:

	// Free buffer
	if (gCfgRomBuf != NULL)
		free (gCfgRomBuf);
	gCfgRomBuf = NULL;

	// Close connection to 1394 device
	if (ghDev != 0)
	{
		MyPrintf1 (VL_MAJOR_STEP, "Closing device %08X... ", gDeviceId);
		ccode = lpfnCls1394CloseHandle (ghDev);
		ghDev = 0;
		if (ccode != TRUE)
			goto DevCloseError;
		MyPrintf (VL_MAJOR_STEP, "OK\n");
	}

	// Unload the drivers
	if (gDriverInitialized == TRUE)
	{
		MyPrintf (VL_MAJOR_STEP, "Unloading drivers ...");
		ccode = lpfnCls1394Terminate ();
		gDriverInitialized = FALSE;
		if (ccode != TRUE)
			goto DriverTerminateError;
		MyPrintf (VL_MAJOR_STEP, "OK\n");
	}

	// Pause before exiting
	if (gStopBeforeExit == TRUE)
	{
		MyPrintf (VL_ALWAYS, "Hit any key to continue.\n");
		getch();
	}
	
	return (0);

CommandLineError:
	MyPrintf(VL_ERROR, "\nERROR:  Command line syntax error\n\n");
	Usage ();
	goto Exit;

LibraryError:
	MyPrintf(VL_ERROR, "\nError: Cannot load cls1394.DLL library and fix run-time pointer to functions\n\n");
	goto Exit;

DriverInitError:
	MyPrintf(VL_ERROR, "\nError: Cannot initialize 1394 drivers\n\n");
	goto Exit;

DevOpenError:
	MyPrintf1(VL_ERROR, "\nError: Device %08X not found\n\n", gDeviceId);
	goto Exit;

AllocError:
	MyPrintf1(VL_ERROR, "\nError: Cannot allocate resource for the configuration ROM buffer\n\n", gDeviceId);
	goto Exit;

ReadError:
	MyPrintf1(VL_ERROR, "\nError: Cannot read the configuration ROM from the device\n\n", gDeviceId);
	goto Exit;

DevCloseError:
	MyPrintf1(VL_ERROR, "\nError: Device %08X cannot be closed properly\n\n", gDeviceId);
	goto Exit;

DriverTerminateError:
	MyPrintf(VL_ERROR, "\nError: Cannot terminate 1394 drivers\n\n");
	goto Exit;

}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ProcessCommandLine
//
//		Processes the command line and sets options and variables accordingly.
//
//	Entry:
//
//		argc			number of arguments in argv[] array
//		argv[]		array of strings containing command line arguments
//
//	Exit:
//
//		TRUE			Arguments parsed and options and variables set.
//		FALSE			Invalid command line detected.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL ProcessCommandLine (int argc, char *argv[])
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	int i;
	int ccode;

	// Assume failure
	retcode = FALSE;

	// Set default values for command line options
	gStopBeforeExit = FALSE;
	gVerbosityLevel = 0;
	gUseSpecifiedOffset = FALSE;
	gSpecifiedOffset = 0;
	gDeviceMode = DM_DEFAULT;
	gDeviceId = DEFAULT_DEVICE_ID;
	gDeviceNumber = DEFAULT_DEVICE_NUMBER;

	// Run through all of the command line options
	// NOTE:  skip the first command line arg.
	// This is the program full path.
	for (i=1; (i<argc) && (*argv[i] != '\0'); i++)
	{
		char *p;
		char c;
		p = argv[i];

		// Parse current arg
		switch (*p)
		{
			// Parse Option	
			case '-':
			case '/':

				// Advance to option
				p++;

				// Parse option character
				switch (toupper(*p++))
				{
					// Verosity Setting
					case 'V':

						// Verify the level number
						sscanf (p, "%c", &c);
						if (c > '9' || c < '0')
							goto Exit;

						// Set the level
						sscanf (p++, "%d", &gVerbosityLevel);
						DBOUT1 (LF_INTERNAL, "gVerbosityLevel = %08X\r\n", gVerbosityLevel);

						// Check if end of option
						if (*p != '\0')
							goto Exit;

						break;

					// Stop Before Exit Option
					case 'S':

						// Set the option
						gStopBeforeExit = TRUE;

						// Check if end of option
						if (*p != '\0')
							goto Exit;

						break;

					// Offset 
					case 'O':

						// Extract the 4 digit specified offset
						gUseSpecifiedOffset = TRUE;
						ccode = sscanf (p, "%04X", &gSpecifiedOffset);
						if (ccode == 0)
							goto Exit;
						DBOUT1 (LF_INTERNAL, "gSpecifiedOffset = %08X\r\n", gSpecifiedOffset);

						break;

					// Device ID
					case 'D':

						// Extract the 8 digit Device ID
						ccode = sscanf (p, "%08X", &gDeviceId);
						if (ccode == 0)
							goto Exit;
						DBOUT1 (LF_INTERNAL, "gDeviceId = %08X\r\n", gDeviceId);

						// Set mode
						gDeviceMode = DM_USE_DEVICE_ID;

						break;

					// Device Number
					case 'N':

						// Set the Device Number
						ccode = sscanf (p++, "%d", &gDeviceNumber);
						if (ccode == 0)
							goto Exit;
						DBOUT1 (LF_INTERNAL, "gDeviceNumber = %08X\r\n", gDeviceNumber);

						// Set mode
						gDeviceMode = DM_USE_DEVICE_NUMBER;

						break;

					// List Device and Choose
					case 'L':

						// Set mode
						gDeviceMode = DM_LIST_DEVICES_AND_CHOOSE;

						// Check if end of option
						if (*p != '\0')
							goto Exit;

						break;

					default:
						goto Exit;
						break;
				}

				break;

			// There's a problem
			default:
				goto Exit;
				break;
		}
	}

	// OK
	retcode = TRUE;

Exit:

	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Usage
//
//		Prints the usage of the program.
//
//	Entry:
//		NONE
//
//	Exit:
//		NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void Usage (void)
{
	MyPrintf (VL_ERROR, "Usage:\n");
	MyPrintf (VL_ERROR, "\n");
	MyPrintf (VL_ERROR, "DumpRom [-n<device number> | -d<device id> | -l] [-v<0-9>] [-s]\n");
	MyPrintf (VL_ERROR, "\n");
	MyPrintf (VL_ERROR, "     -n<device number> device number, 0 - 63\n");
	MyPrintf (VL_ERROR, "                       Default: the first device on the bus\n");
	MyPrintf (VL_ERROR, "\n");
	MyPrintf (VL_ERROR, "     -d<device id>     32-bit hexadecimal device ID = node_vendor_id+chip_id_lo\n");
	MyPrintf (VL_ERROR, "                       Default: ID of the first device on the bus\n");
	MyPrintf (VL_ERROR, "\n");
	MyPrintf (VL_ERROR, "     -l                List all peripheral device on the bus and allow the user to chose one\n");
	MyPrintf (VL_ERROR, "\n");
	MyPrintf (VL_ERROR, "     -v<0-9>           Set verbosity level\n");
	MyPrintf (VL_ERROR, "                       0 = none\n");
	MyPrintf (VL_ERROR, "                       1 = show major steps\n");
	MyPrintf (VL_ERROR, "                       2 = show major steps and debug information\n");
	MyPrintf (VL_ERROR, "\n");
	MyPrintf (VL_ERROR, "     -o<1394 offset>   Dump only the quadlet at the specified 4 digit\n");
	MyPrintf (VL_ERROR, "                       configuration rom offset.  Offset added to base\n");
	MyPrintf (VL_ERROR, "                       of 0xXXXXFFFF:F0000400.\n");
	MyPrintf (VL_ERROR, "                       Default: dumps the whole configuration rom\n");
	MyPrintf (VL_ERROR, "\n");
	MyPrintf (VL_ERROR, "     -s                Stop before exiting\n");
	MyPrintf (VL_ERROR, "\n");
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPrintf
//
//		My own special brew of printf.  Allows a command line verbosity level
//		to determine when a message is displayed on the screen or debug terminal.
//
//	Entry:
//
//		vl			verbosity level
//
//		fmt		printf format string.
//
//	Exit:
//		NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPrintf (int vl, char *fmt)
{
	//#define DEBUGFLAGS DebugThisRoutine
	if (vl <= gVerbosityLevel)
	{
		DBOUT (DB_ALWAYS, fmt);
		printf (fmt);
	}
	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPrintf1
//
//		My own special brew of printf with 1 DWORD input argument.
//		Allows a command line verbosity level to determine when a 
//		message is displayed on the screen or debug terminal.
//
//	Entry:
//
//		vl			verbosity level
//
//		fmt		printf format string.
//
//		p1			DWORD parameter 1.  Can be anything (e.g. ptr to string, etc...)
//
//	Exit:
//		NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPrintf1 (int vl, char *fmt, DWORD p1)
{
	char msg[500];
	sprintf (msg, fmt, p1);
	MyPrintf (vl, msg);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPrintf2
//
//		My own special brew of printf with 2 DWORD input arguments.
//		Allows a command line verbosity level to determine when a 
//		message is displayed on the screen or debug terminal.
//
//	Entry:
//
//		vl			verbosity level
//
//		fmt		printf format string.
//
//		p1			DWORD parameter 1.  Can be anything (e.g. ptr to string, etc...)
//
//		p2			DWORD parameter 2.  Can be anything (e.g. ptr to string, etc...)
//
//	Exit:
//		NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPrintf2 (int vl, char *fmt, DWORD p1, DWORD p2)
{
	char msg[500];
	sprintf (msg, fmt, p1, p2);
	MyPrintf (vl, msg);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPrintf3
//
//		My own special brew of printf with 3 DWORD input arguments.
//		Allows a command line verbosity level to determine when a 
//		message is displayed on the screen or debug terminal.
//
//	Entry:
//
//		vl			verbosity level
//
//		fmt		printf format string.
//
//		p1			DWORD parameter 1.  Can be anything (e.g. ptr to string, etc...)
//
//		p2			DWORD parameter 2.  Can be anything (e.g. ptr to string, etc...)
//
//		p3			DWORD parameter 3.  Can be anything (e.g. ptr to string, etc...)
//
//	Exit:
//		NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPrintf3 (int vl, char *fmt, DWORD p1, DWORD p2, DWORD p3)
{
	char msg[500];
	sprintf (msg, fmt, p1, p2, p3);
	MyPrintf (vl, msg);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPrintf4
//
//		My own special brew of printf with 4 DWORD input arguments.
//		Allows a command line verbosity level to determine when a 
//		message is displayed on the screen or debug terminal.
//
//	Entry:
//
//		vl			verbosity level
//
//		fmt		printf format string.
//
//		p1			DWORD parameter 1.  Can be anything (e.g. ptr to string, etc...)
//
//		p2			DWORD parameter 2.  Can be anything (e.g. ptr to string, etc...)
//
//		p3			DWORD parameter 3.  Can be anything (e.g. ptr to string, etc...)
//
//		p4			DWORD parameter 4.  Can be anything (e.g. ptr to string, etc...)
//
//	Exit:
//		NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPrintf4 (int vl, char *fmt, DWORD p1, DWORD p2, DWORD p3, DWORD p4)
{
	char msg[500];
	sprintf (msg, fmt, p1, p2, p3, p4);
	MyPrintf (vl, msg);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPrintf5
//
//		My own special brew of printf with 5 DWORD input arguments.
//		Allows a command line verbosity level to determine when a 
//		message is displayed on the screen or debug terminal.
//
//	Entry:
//
//		vl			verbosity level
//
//		fmt		printf format string.
//
//		p1			DWORD parameter 1.  Can be anything (e.g. ptr to string, etc...)
//
//		p2			DWORD parameter 2.  Can be anything (e.g. ptr to string, etc...)
//
//		p3			DWORD parameter 3.  Can be anything (e.g. ptr to string, etc...)
//
//		p4			DWORD parameter 4.  Can be anything (e.g. ptr to string, etc...)
//
//		p5			DWORD parameter 5.  Can be anything (e.g. ptr to string, etc...)
//
//	Exit:
//		NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPrintf5 (int vl, char *fmt, DWORD p1, DWORD p2, DWORD p3, DWORD p4, DWORD p5)
{
	char msg[500];
	sprintf (msg, fmt, p1, p2, p3, p4, p5);
	MyPrintf (vl, msg);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPrintf6
//
//		My own special brew of printf with 6 DWORD input arguments.
//		Allows a command line verbosity level to determine when a 
//		message is displayed on the screen or debug terminal.
//
//	Entry:
//
//		vl			verbosity level
//
//		fmt		printf format string.
//
//		p1			DWORD parameter 1.  Can be anything (e.g. ptr to string, etc...)
//
//		p2			DWORD parameter 2.  Can be anything (e.g. ptr to string, etc...)
//
//		p3			DWORD parameter 3.  Can be anything (e.g. ptr to string, etc...)
//
//		p4			DWORD parameter 4.  Can be anything (e.g. ptr to string, etc...)
//
//		p5			DWORD parameter 5.  Can be anything (e.g. ptr to string, etc...)
//
//		p6			DWORD parameter 6.  Can be anything (e.g. ptr to string, etc...)
//
//	Exit:
//		NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPrintf6 (int vl, char *fmt, DWORD p1, DWORD p2, DWORD p3, DWORD p4, DWORD p5, DWORD p6)
{
	char msg[500];
	sprintf (msg, fmt, p1, p2, p3, p4, p5, p6);
	MyPrintf (vl, msg);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// WriteQuadlet
//
//		Writes a quadlet to the specified device at the specified 1394 address.
//
//	Entry:
//
//		hDev			handle to 1394 device
//
//		liAddress	1394 address to the quadlet to write
//
//		qValue		quadlet to be written.
//
//	Exit:
//
//		TRUE			Operation was successful.
//
//		FALSE			Operation failed.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL WriteQuadlet (
			cls1394HANDLE hDev,
			LARGE_INTEGER liAddress,
			QUADLET qValue)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	BOOL ccode;
	clsAsyncWrite aw;

	DBOUT (LF_ENTRY, "WriteQuadlet\r\n");
	DBOUT2 (LF_ENTRY_PARAMS, "liAddress = %08X%08X\r\n", liAddress.HighPart, liAddress.LowPart);
	DBOUT1 (LF_ENTRY_PARAMS, "qValue = %08X\r\n", qValue);

	// Assume failure
	retcode = FALSE;

	// Setup async write structure
	aw.DestinationAddress = &liAddress;
	aw.nNumberOfBytesToWrite = sizeof(QUADLET);
	aw.nBlockSize = 0;
	aw.fulFlags = 0;
	aw.lpBuffer = &qValue;

	// Call the driver to do the async write
	ccode = lpfnCls1394DeviceIoControl(
				hDev,
				CLS_REQUEST_ASYNC_WRITE,
				&aw,
				sizeof(clsAsyncWrite),
				0,
				0,
				0,
				0);
	if (ccode != TRUE)
		goto Exit;

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "WriteQuadlet, retcode = %08X\r\n", retcode);
	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// WriteQuadletBuffer
//
//		Writes an array of quadlets to the specified device at the 
//		specified 1394 address.
//
//	Entry:
//
//		hDev			handle to 1394 device
//
//		liAddress	1394 address to the quadlet to write
//
//		Buffer		array of quadlets to be written.
//
//		BufferSize	size of Buffer in bytes
//
//	Exit:
//
//		TRUE			Operation was successful.
//
//		FALSE			Operation failed.
//
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL WriteQuadletBuffer (
			cls1394HANDLE hDev,
			LARGE_INTEGER liAddress,
			PQUADLET Buffer,
			DWORD BufferSize)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	BOOL ccode;
	clsAsyncWrite aw;

	DBOUT (LF_ENTRY, "WriteQuadletBuffer\r\n");
	DBOUT1 (LF_ENTRY_PARAMS, "hDev = %08X\r\n", hDev);
	DBOUT2 (LF_ENTRY_PARAMS, "liAddress = %08X%08X\r\n", liAddress.HighPart, liAddress.LowPart);
	DBOUT1 (LF_ENTRY_PARAMS, "Buffer = %08X\r\n", Buffer);
	DBOUT1 (LF_ENTRY_PARAMS, "BufferSize = %08X\r\n", BufferSize);

	// Assume failure
	retcode = FALSE;

	// Setup async write structure
	aw.DestinationAddress = &liAddress;
	aw.nNumberOfBytesToWrite = BufferSize;
	aw.nBlockSize = 0;
	aw.fulFlags = 0;
	aw.lpBuffer = Buffer;

	// Call the driver to do the async write
	ccode = lpfnCls1394DeviceIoControl(
				hDev,
				CLS_REQUEST_ASYNC_WRITE,
				&aw,
				sizeof(clsAsyncWrite),
				0,
				0,
				0,
				0);
	if (ccode != TRUE)
		goto Exit;

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "WriteQuadletBuffer, retcode = %08X\r\n", retcode);
	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// WriteVerifyQuadlet
//
//		Writes a quadlet to the specified device at the specified 1394 address.
//		Then reads back the quadlet and performs a comparison to verify that
//		the write succeeded.
//
//	Entry:
//
//		hDev			handle to 1394 device
//
//		liAddress	1394 address to the quadlet to write
//
//		qValue		quadlet to be written.
//
//		to				timeout in milliseconds.
//
//	Exit:
//
//		TRUE			Operation was successful.
//
//		FALSE			Operation failed.
//
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL WriteVerifyQuadlet (
			cls1394HANDLE hDev,
			LARGE_INTEGER liAddress,
			QUADLET qValue,
			DWORD to)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	BOOL ccode;
	QUADLET qVerify;
	DWORD tc;

	DBOUT (LF_ENTRY, "WriteVerifyQuadlet\r\n");
	DBOUT2 (LF_ENTRY_PARAMS, "liAddress = %08X%08X\r\n", liAddress.HighPart, liAddress.LowPart);
	DBOUT1 (LF_ENTRY_PARAMS, "qValue = %08X\r\n", qValue);

	// Assume failure
	retcode = FALSE;

	// Write the quadlet
	ccode = WriteQuadlet (hDev, liAddress, qValue);
	if (ccode != TRUE)
		goto Exit;

	// Verify the quadlet
	// Wait for the completion with a timeout
	tc = GetTickCount();
	do
	{
		// Read the quadlet
		qVerify = ~qValue;
		ccode = ReadQuadlet (hDev, liAddress, &qVerify);
		if (ccode != TRUE)
			continue;

		// Verify the QUADLET
		if (qValue == qVerify)
			goto OK;

	} while ((tc + to) > GetTickCount());

	// Timed out -- FAILED!
	goto Exit;

OK:

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "WriteVerifyQuadlet, retcode = %08X\r\n", retcode);
	return (retcode);

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ReadQuadlet
//
//		Writes a quadlet to the specified device at the specified 1394 address.
//		Then reads back the quadlet and performs a comparison to verify that
//		the write succeeded.
//
//	Entry:
//
//		hDev			handle to 1394 device
//
//		liAddress	1394 address to the quadlet to write
//
//		qValue		pointer to quadlet to be filled in.
//
//	Exit:
//
//		TRUE			Operation was successful.
//
//		FALSE			Operation failed.
//
//		*qValue		if function returns TRUE, this is the quadlet value read
//						from the specified device and address.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL ReadQuadlet(cls1394HANDLE hDev, LARGE_INTEGER liAddress, QUADLET *qValue)
{
    clsAsyncRead aw;
    BOOL rc;

	// Setup request structure
    aw.DestinationAddress = &liAddress;
    aw.nNumberOfBytesToRead = sizeof(QUADLET);
    aw.nBlockSize = 0;
    aw.fulFlags = 0;
    aw.lpBuffer = qValue;

	// Call driver
    rc = lpfnCls1394DeviceIoControl(
				hDev,
           CLS_REQUEST_ASYNC_READ,
           &aw,
           sizeof(clsAsyncRead),
           0,
           0,
           0,
           0);
	return (rc);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DeviceOpen
//
//		Opens a 1394 device according to the mode, device number, device ID
//		specified.
//
//	Entry:
//
//		DeviceMode		specifies how to determine which device to open.
//							
//							DM_USE_DEVICE_NUMBER indicates to use the DeviceNumber
//							parameter only.
//							
//							DM_USE_DEVICE_ID indicates to use the DeviceId
//							parameter only.
//							
//							DM_LIST_DEVICES_AND_CHOOSE indicates to list all of the
//							1394 devices and allow the user to choose a device number.
//							Once a choice is made, that device number is used to
//							attempt to open the device.
//
//		DeviceNumber	Number of the device to open.  Can only be used with
//							DeviceMode = DM_USE_DEVICE_NUMBER.
//
//							0 = local 1394 host adapter (TSBKPCI "TI Lynx").
//							1 and above = one of the 1394 perpheral connected to the TSBKPCI.
//
//		DeviceId			1394 Device ID to open.  Can only be used with 
//							DeviceMode = DM_USE_DEVICE_ID.
//							The device ID is usually the 1394 Vendor ID of the device
//							which the TSBKPCI read from the device's configuration ROM.
//
//		hDev				pointer to a handle to be filled in.
//
//	Exit:
//
//		TRUE				Operation succeeded.
//
//		FALSE				Operation failed.
//
//		*hDev				if the return value is TRUE, the is the handle to the device opened.
//						
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL DeviceOpen (DWORD DeviceMode, PDWORD DeviceNumber, PDWORD DeviceId, cls1394HANDLE *hDev)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	cls1394HANDLE lochDev;
	DeviceInfo di;
	WORD locDevNum;
	DWORD locDevId;
	BOOL ccode;
	DWORD numDevices;
	DWORD i;
	int ccode2;
	int choice;

	DBOUT (LF_ENTRY, "DeviceOpen\r\n");

	// Assume failure
	retcode = FALSE;
	lochDev = 0;
	locDevNum = (WORD)(*DeviceNumber);
	locDevId = *DeviceId;

	// What device mode are we in?
	switch (DeviceMode)
	{
		case DM_USE_DEVICE_NUMBER:

			// Get the specified device on the bus
			ccode = lpfnGetDeviceInfo(locDevNum, &di);
			if (ccode != TRUE)
				goto DeviceInfoError;
			locDevId = di.VendID_DevType;

			break;

		case DM_USE_DEVICE_ID:

			// Do we need to get the first device on the bus?
			if (locDevId == DEFAULT_DEVICE_ID)
			{
				locDevNum = 1;
				ccode = lpfnGetDeviceInfo(locDevNum, &di);
				if (ccode != TRUE)
					goto DeviceInfoError;
				locDevId = di.VendID_DevType;
			}

			break;

		case DM_LIST_DEVICES_AND_CHOOSE:
		default:

			// List all devices on the bus including the host adapter
			MyPrintf2(VL_ALWAYS, "Num  DeviceID\n", i, di.VendID_DevType);
			MyPrintf (VL_ALWAYS, "---  --------\n");
			for (i=0; 1; i++)
			{
				ccode = lpfnGetDeviceInfo ((WORD)i, &di);
				if (ccode != TRUE)
					break;
				MyPrintf2 (VL_ALWAYS, "%3d  %08X\n", i, di.VendID_DevType);
			}
			numDevices = i;
			if (numDevices == 0)
				goto NoDevicesOnBus;

			// Prompt the user to choose a device
			do
			{
				MyPrintf1 (VL_ALWAYS, "Please select one of the above (0 - %d): ", numDevices-1);
				ccode2 = fscanf (stdin, "%d", &choice);
			} while ((ccode2 == 0) || !((choice >= 0) && ((DWORD)choice <= numDevices)));

			// Get information on the chosen device
			locDevNum = choice;
			ccode = lpfnGetDeviceInfo(locDevNum, &di);
			if (ccode != TRUE)
				goto DeviceInfoError;
			locDevId = di.VendID_DevType;

			break;
	}

	// Try to open the device
	DBOUT1 (LF_INTERNAL, "Opening device with ID = %08X\r\n", locDevId);
	if (locDevNum == 0)
		lochDev = lpfnCls1394CreateFile (locDevId, 0);
	else
		lochDev = lpfnCls1394CreateFile (locDevId, 1);
	if (lochDev == 0)
		goto OpenError;

	// OK
	retcode = TRUE;
	*hDev = lochDev;
	*DeviceId = locDevId;
	*DeviceNumber = (DWORD)locDevNum;
	DBOUT1 (LF_EXIT_PARAMS, "*hDev = %08X\r\n", *hDev);

Exit:

	DBOUT1 (LF_EXIT, "DeviceOpen, retcode = %08X\r\n", retcode);
	return (retcode);

DeviceInfoError:
	MyPrintf1 (VL_ERROR, "\n\nERROR:  Cannot retrieve information for device number %d\n\n", locDevNum);
	goto Exit;

OpenError:
	MyPrintf1 (VL_ERROR, "\n\nERROR:  Cannot open device ID %08X\n\n", locDevId);
	goto Exit;

NoDevicesOnBus:
	MyPrintf (VL_ERROR, "\n\nERROR:  There are no peripheral devices on the 1394 bus\n\n");
	goto Exit;

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// AllocateConfigRomBuffer
//
//		Allocates a quadlet buffer to fit the entire configuration ROM of
//		the specified device.  This routine will read the first quadlet
//		of the configuration ROM and allocate the number of quadlets
//		specified in the second byte.  It will also take minimal 
//		configuration ROMs into consideration.
//
//	Entry:
//
//		hDev				handle of the device
//
//		CfgRomBuf		pointer buffer pointer varaible to be filled in
//
//		CfgRomBufSize	pointer to size variable to be filled in
//
//	Exit:
//
//		TRUE				Operation succeeded
//
//		FALSE				Opearation failed
//
//		*cfgRomBuf		if return value = TRUE, points to an allocated quadlet buffer.

//		*cfgRomBufSize	if return value = TRUE, number of bytes in cfgRomBuf.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL AllocateConfigRomBuffer (cls1394HANDLE hDev, PQUADLET *CfgRomBuf, PDWORD CfgRomBufSize)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	LARGE_INTEGER liAddress;
	QUADLET qValue;
	BOOL ccode;
	DWORD numCfgRomQuads;
	DWORD romBufSize;
	PQUADLET romBuf;

	DBOUT (LF_ENTRY, "AllocateConfigRomBuffer\r\n");
	DBOUT1 (LF_ENTRY_PARAMS, "hDev = %08X\r\n", hDev);

	// Assume failure
	retcode = FALSE;
	*CfgRomBuf = NULL;
	*CfgRomBufSize = 0;
	romBuf = NULL;

	// Read the first quadlet of the CFG ROM
	liAddress = CONFIG_ROM;
	ccode = ReadQuadlet(hDev, liAddress, &qValue);
	if (ccode != TRUE)
		goto ReadQuadError;

	// First quadlet is in either one of the following forms:
	//
	//	01VVVVVV
	//		Minimal CFG ROM.
	//		VVVVVV = Vendor ID
	// 
	//	04NNCCCC
	//		General CFG ROM.
	//		04 =   Number of quadlets in the Bus_Info_Block
	//		NN =   Number of quadlets covered by the CRC 
	//		       (i.e.  number of quadlets in the CFG ROM, excluding the first quadlet)
	//		CCCC = CRC-16 of the CFG ROM (excluding the first quadlet)
	switch ((qValue & 0xFF000000)>>24)
	{
		case 0x01:
			// Minimal cfg rom, only 1 quadlet
			numCfgRomQuads = 1;
			break;
		case 0x04:
			// Extract the number of quadlets in the cfg rom
			numCfgRomQuads = ((qValue &0x00FF0000) >> 16) + 1;
			break;
		default:
			// Invalid config rom
			goto InvalidCfgRomError;
			break;
	}

	// Allocate the buffer
	romBufSize = numCfgRomQuads * sizeof(QUADLET);
	romBuf = malloc (romBufSize);
	if (romBuf == NULL)
		goto AllocError;

	// OK
	retcode = TRUE;
	*CfgRomBuf = romBuf;
	*CfgRomBufSize = romBufSize;
	DBOUT1 (LF_EXIT_PARAMS, "DumpConfigRom, retcode = %08X\r\n", retcode);
	DBOUT1 (LF_EXIT_PARAMS, "DumpConfigRom, retcode = %08X\r\n", retcode);

Exit:

	DBOUT1 (LF_EXIT, "DumpConfigRom, retcode = %08X\r\n", retcode);
	return (retcode);

ReadQuadError:
	MyPrintf2 (VL_ERROR, "\n\nERROR:  Cannot read quadlet from address offset %04X:%08X\n\n", liAddress.HighPart, liAddress.LowPart);
	goto Exit;

InvalidCfgRomError:
	MyPrintf1 (VL_ERROR, "\n\nERROR:  Invalid configuration ROM, header quadlet = %08X\n\n", qValue);
	goto Exit;

AllocError:
	MyPrintf1 (VL_ERROR, "\n\nERROR:  Cannot allocate %d bytes of memory for the configuration ROM buffer\n\n", romBufSize);
	goto Exit;

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ReadConfigRomBuffer
//
//		Reads the configuration ROM of the specified device into the specified
//		quadlet buffer.
//
//	Entry:
//
//		hDev				handle to the device
//
//		CfgRomBuf		pointer to a quadlet buffer to filled in.
//
//		CfgRomBufSize	size in bytes of CfgRomBuf.
//
//	Exit:
//
//		TRUE				Operation succeeded
//
//		FALSE				Operation failed
//
//		CfgRomBuf		If the return value = TRUE, contains the image
//							of the configruation ROM.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL ReadConfigRomBuffer (cls1394HANDLE hDev, PQUADLET CfgRomBuf, DWORD CfgRomBufSize)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	LARGE_INTEGER liAddress;
	BOOL ccode;
	PQUADLET pq;

	DBOUT (LF_ENTRY, "ReadConfigRomBuffer\r\n");
	DBOUT1 (LF_ENTRY_PARAMS, "hDev %08X\r\n", hDev);

	// Assume failure
	retcode = FALSE;

	// Just read until we either get an error or hit end of buffer
	for (	liAddress = CONFIG_ROM, pq = CfgRomBuf; 
			liAddress.LowPart < CONFIG_ROM.LowPart+CfgRomBufSize; 
			liAddress.LowPart+=4, pq++)
	{
		// Read the next quadlet of the CFG ROM
		ccode = ReadQuadlet(hDev, liAddress, pq);
		if (ccode != TRUE)
			goto ReadQuadError;
	}

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "ReadConfigRomBuffer, retcode = %08X\r\n", retcode);
	return (retcode);

ReadQuadError:
	MyPrintf2 (VL_ERROR, "\n\nERROR:  Cannot read quadlet from address offset %04X:%08X\n\n", liAddress.HighPart, liAddress.LowPart);
	goto Exit;

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DumpSpecificQuadlet
//
//		Reads the value of the quadlet at the specified configuration ROM offset
//		on the specified device and then prints it to the display.
//
//	Entry:
//
//		hDev					handle to the device
//
//		SpecifiedOffset	offset from the beginning of the 1394 configuration
//								ROM address offset (0xFFFF:F0000400)
//
//	Exit:
//
//		TRUE					operation succeeded
//
//		FALSE					operation failed
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL DumpSpecificQuadlet (cls1394HANDLE hDev, DWORD SpecifiedOffset)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	LARGE_INTEGER liAddress;
	BOOL ccode;
	QUADLET qValue;

	// Assume failure
	retcode = FALSE;

	// Read the quadlet at the specified cfg rom offset
	qValue = 0;
	liAddress = CONFIG_ROM;
	liAddress.LowPart += gSpecifiedOffset;
	ccode = ReadQuadlet(hDev, liAddress, &qValue);
	if (ccode != TRUE)
		goto ReadQuadError;

	// Print it
	MyPrintf4(VL_ALWAYS, "\nQuadlet at address offset %04X:%08X (%04X) = %08X\n", 
		CONFIG_ROM.HighPart, CONFIG_ROM.LowPart+gSpecifiedOffset, gSpecifiedOffset, qValue);

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "ReadConfigRomBuffer, retcode = %08X\r\n", retcode);
	return (retcode);

ReadQuadError:
	MyPrintf2 (VL_ERROR, "\n\nERROR:  Cannot read quadlet from address offset %04X:%08X\n\n", liAddress.HighPart, liAddress.LowPart);
	goto Exit;

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DumpConfigRom
//
//		Displays the raw quadlet data contained in the specified 
//		configuration ROM buffer.
//
//	Entry:
//
//		CfgRomBuf		quadlet buffer containing the configuration ROM image
//
//		CfgRomBufSize	size in bytes of CFGRomBuf
//
//	Exit:
//
//		TRUE				operation succeeded
//
//		FALSE				operation failed
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL DumpConfigRom (PQUADLET CfgRomBuf, DWORD CfgRomBufSize)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;
	DWORD i;
	PQUADLET pq;

	DBOUT (LF_ENTRY, "DumpConfigRom\r\n");

	// Assume failure
	retcode = FALSE;

	// Check if the size is in quadlets
	if (CfgRomBufSize % sizeof(QUADLET))
		goto InvalidSize;

	// RAW Dump of buffer
	MyPrintf (VL_ALWAYS, "\n");
	MyPrintf2 (VL_ALWAYS, "Raw Data Dump of the Configuration ROM from device number %d (id=%08X)\n\n", gDeviceNumber, gDeviceId);
	MyPrintf (VL_ALWAYS, "1394 Addr Off    Data\n");
	MyPrintf (VL_ALWAYS, "-------------    --------  --------  --------  -------- ");
	for (i=0, pq=CfgRomBuf; i<CfgRomBufSize; i+=4, pq++)
	{
		// Display the quadlet
		if ((i % 16) == 0)
		{
			MyPrintf2 (VL_ALWAYS, "\n%04X:%08X    ", CONFIG_ROM.HighPart, CONFIG_ROM.LowPart+i);
		}
		MyPrintf1 (VL_ALWAYS, "%08X  ", *pq);
	}
	MyPrintf (VL_ALWAYS, "\n");

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "DumpConfigRom, retcode = %08X\r\n", retcode);
	return (retcode);

InvalidSize:
	MyPrintf2 (VL_ERROR, "\n\nERROR:  Buffer size %d is not divisible by %d\n\n", CfgRomBufSize, sizeof(QUADLET));
	goto Exit;

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DecodeConfigRom
//
//		Decodes the specified configuration ROM buffer according to the 
//		1212, 1394 and SBP-2 specifications.
//
//	Entry:
//
//		CfgRomBuf		quadlet buffer containing the configuration ROM image
//
//		CfgRomBufSize	size in bytes of CFGRomBuf
//
//	Exit:
//
//		TRUE				operation succeeded
//
//		FALSE				operation failed
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL DecodeConfigRom (PQUADLET CfgRomBuf, DWORD CfgRomBufSize)
{
	//#define DEBUGFLAGS DebugThisRoutine

	// Block Types
	#define BT_NONE			0
	#define BT_DIRECTORY		1
	#define BT_LEAF			2

	// Directory Types
	// NOTE:  Use the 1212 key_value definitions (e.g. Unit_Directory) and
	// define the directory types not defined in 1212
	#define DT_UNKNOWN		0
	#define DT_ROOT			-1

	// Leaf Types
	// NOTE:  Use the 1212 key_value definitions (e.g. Textual_Descriptor) and
	// define the directory types not defined in 1212
	#define LT_UNKNOWN		0

	// Block Information Structure
	//
	// For every quadlet entry in the CFG ROM there will be a corresponding entry
	// in the Block Information Table.
	//
	// Used to determine if the given CFG ROM entry is the beginning of 
	// a new block.  The BlockType will be BT_NONE if the entry is just an
	// entry in the currrent block.  Otherwise it will specify what type 
	// of block it is.  BT_DIRECTORY indicates that this is a directory block.
	// BT_LEAF indicates that this is a leaf block.
	//
	//	ReferringDirEntry is a copy of the quadlet (directory entry) in a previous block
	// that refers to this block.
	//
	// ReferringDirEntryIdx the the index of the directory entry in a previous block
	// that refered to this block.  Need to this to show where this was referred from.
	//
	typedef struct BlockInfoStruct
	{
		DWORD			BlockType;
		DirEntryT	ReferringDirEntry;
		DWORD			ReferringDirEntryIdx;
	} BlockInfoT, *PBlockInfoT;

	// Helper macros for repeated operations.  Could make these functions, but
	// I like to hear the purists gripe.  Plus it's easier to change if they
	// are defined right here.
	#define INC_PLACE_HOLDERS {cr++; bi++; curTabIdx++;}
	#define INC_PLACE_HOLDERS2 (cr++, bi++, curTabIdx++)
	#define PRINT_ADDR_OFF_VALUE MyPrintf3 (VL_ALWAYS, "%04X:%08X %08X ", CONFIG_ROM.HighPart, CONFIG_ROM.LowPart+(curTabIdx*sizeof(QUADLET)), *cr);
	#define CALC_LOW_PART_CFG_ROM_INDIRECT_ADDR_OFF(io) CONFIG_ROM.LowPart+(((curTabIdx+io)<maxTabIdx)?((curTabIdx+io)*sizeof(QUADLET)):0x0FFFFFFF)
	#define CALC_LOW_PART_CSR_BASE_ADDR_OFF(co) CSR_BASE.LowPart+(co*sizeof(QUADLET))
	#define SET_BLOCK_INFO_INDIRECT(io,bt,de,dei) if((bi+io)<biEnd){bi[io].BlockType=bt;bi[io].ReferringDirEntry.q=de;bi[io].ReferringDirEntryIdx=dei;}
	#define PRINT_HEADER_INFO MyPrintf2 (VL_ALWAYS, "length=%04X, crc=%04X", ((PBlockHeaderT)cr)->length, ((PBlockHeaderT)cr)->crc);

	BOOL retcode;
	DWORD curTabIdx;
	DWORD maxTabIdx;
	PQUADLET cr;
	PQUADLET crEnd;
	PBlockInfoT blockInfo;
	PBlockInfoT bi;
	PBlockInfoT biEnd;
	PCfgRomHeaderT romHdr;
	DWORD curBlockType;
	PDirEntryT dirEntry;
	DWORD curUnitSpecId;
	DWORD curUnitSwVersion;

	DBOUT (LF_ENTRY, "DecodeConfigRom\r\n");

	// Assume failure
	retcode = FALSE;

	// Init vars
	curTabIdx = 0;
	maxTabIdx = CfgRomBufSize/sizeof(QUADLET);
	blockInfo = NULL;
	cr = CfgRomBuf;
	crEnd = cr + maxTabIdx;
	bi = NULL;
	biEnd = NULL;
	curUnitSpecId = 0;
	curUnitSwVersion = 0;

	// Check if the size is in quadlets
	if (CfgRomBufSize % sizeof(QUADLET))
		goto InvalidSize;

	// Allocate a helper block type array which has a 1:1 correspondence
	// to each quadlet in the CFG ROM.
	// NOTE:  because the CFG ROM is hierarchical, we must
	// save information somewhere about blocks that are
	// referred to from other blocks during parsing.  Otherwise, when
	// we get to that block, we will not be able to distinquish the 
	// block's type and how to decode it.
	blockInfo = calloc (sizeof(BlockInfoT), maxTabIdx);
	if (blockInfo == NULL)
		goto AllocError;
	bi = blockInfo;
	biEnd = bi + maxTabIdx;

	// Print header
	MyPrintf (VL_ALWAYS, "\n");
	MyPrintf2 (VL_ALWAYS, "Decode of the Configuration ROM from device number %d (id=%08X)\n\n", gDeviceNumber, gDeviceId);
	MyPrintf (VL_ALWAYS, "1394 Addr Off Quadlet  Meaning\n");
	MyPrintf (VL_ALWAYS, "------------- -------- --------------------------------------------------\n");

	// Is this a minimal CFG ROM?
	romHdr = (PCfgRomHeaderT)cr;
	if (romHdr->mcr.info_length == 0x01)
		goto MinimalCfgRom;

	// Is this a valid general 1394 CFG ROM?
	if (romHdr->gcr.info_length != 0x04)
		goto InvalidRomHdr;

	// Display CFG ROM Header
	MyPrintf (VL_ALWAYS, "\nConfiruation ROM Header\n");
	PRINT_ADDR_OFF_VALUE
	MyPrintf3 (VL_ALWAYS, "info_length=%02X, crc_length=%02X, rom_crc_value=%04X\n", 
		romHdr->gcr.info_length, romHdr->gcr.crc_length, romHdr->gcr.rom_crc_value);
	INC_PLACE_HOLDERS

	// Check the bus_name in the bus_info_block
	if (romHdr->gcr.bus_info_block.bus_name != BUS_NAME_1394)
		goto InvalidRomHdr;

	// Display bus_info_block
	MyPrintf (VL_ALWAYS, "\nBus_Info_Block\n");
	PRINT_ADDR_OFF_VALUE
	MyPrintf1 (VL_ALWAYS, "bus_name=%08X (\"", romHdr->gcr.bus_info_block.bus_name);
	DumpQuadletAsChars (romHdr->gcr.bus_info_block.bus_name);
	MyPrintf (VL_ALWAYS, "\")\n");
	INC_PLACE_HOLDERS
	PRINT_ADDR_OFF_VALUE
	MyPrintf6 (VL_ALWAYS, "irmc=%X, cmc=%X, isc=%X, bmc=%X, cyc_clk_acc=%02X, max_rec=%X\n",
		romHdr->gcr.bus_info_block.irmc,
		romHdr->gcr.bus_info_block.cmc,
		romHdr->gcr.bus_info_block.isc,
		romHdr->gcr.bus_info_block.bmc,
		romHdr->gcr.bus_info_block.cyc_clk_acc,
		romHdr->gcr.bus_info_block.max_rec);
	INC_PLACE_HOLDERS
	PRINT_ADDR_OFF_VALUE
	MyPrintf2 (VL_ALWAYS, "node_vendor_id=%06X, chip_id_hi=%02X\n",
		romHdr->gcr.bus_info_block.node_vendor_id,
		romHdr->gcr.bus_info_block.chip_id_hi);
	INC_PLACE_HOLDERS
	PRINT_ADDR_OFF_VALUE
	MyPrintf1 (VL_ALWAYS, "chip_id_lo=%08X\n", romHdr->gcr.bus_info_block.chip_id_lo);
	INC_PLACE_HOLDERS

	// We should be at the begining of the root_directory.
	// It is possible that the CFG ROM consist only of
	// a ROM Header and BIB.
	if (cr >= crEnd)
		goto OK;

	// Decode the rest of the CFG ROM starting at the root_directory
	for (	bi->BlockType = BT_DIRECTORY, bi->ReferringDirEntry.q = DT_ROOT, curBlockType = BT_DIRECTORY;
			cr < crEnd;
			INC_PLACE_HOLDERS2)
	{
		// Are we at the start of a new block?
		switch (bi->BlockType)
		{
			// It's a element of a block
			case BT_NONE:

				// Print the current address and quadlet value
				PRINT_ADDR_OFF_VALUE

				// Are currently parsing a directory?
				// If so, we need to parse the directory entry
				if (curBlockType == BT_DIRECTORY)
				{
					// Process according to the key_type
					dirEntry = (PDirEntryT)cr;
					//MyPrintf1 (VL_ALWAYS, "key=%02X ", dirEntry->general.key);
					MyPrintf1 (VL_ALWAYS, "%s ", 
						(DWORD)KeyToString(
								dirEntry, 
								curUnitSpecId, 
								curUnitSwVersion));
					switch (dirEntry->iv.key_type)
					{
						case KT_IMMEDIATE_VALUE:

							// Remember the last Unit_Spec_Id and Unit_Sw_Version
							// Used to differentiate protocol specific key_values
							if (dirEntry->iv.key_value == Unit_Spec_Id)
								curUnitSpecId = dirEntry->iv.value;
							else if (dirEntry->iv.key_value == Unit_Sw_Version)
								curUnitSwVersion = dirEntry->iv.value;

							// Decode the immediate value
							MyPrintf1 (VL_ALWAYS, "%s", 
								(DWORD)ImmediateValueToString
										(dirEntry, 
										curUnitSpecId, 
										curUnitSwVersion));

							break;

						case KT_CSR_OFFSET:
							MyPrintf3 (VL_ALWAYS, "crc_offset=%06X (%04X:%08X)", 
								dirEntry->co.csr_offset, 
								CSR_BASE.HighPart, 
								CALC_LOW_PART_CSR_BASE_ADDR_OFF(dirEntry->co.csr_offset));
							break;

						case  KT_LEAF_INDIRECT_OFFSET:
							MyPrintf3 (VL_ALWAYS, "ind_off=%06X (%04X:%08X)", 
								dirEntry->io.indirect_offset, 
								CONFIG_ROM.HighPart, 
								CALC_LOW_PART_CFG_ROM_INDIRECT_ADDR_OFF(dirEntry->io.indirect_offset));

							// Set the referenced block info table entry to LEAF type
							SET_BLOCK_INFO_INDIRECT(dirEntry->io.indirect_offset,BT_LEAF,dirEntry->q,curTabIdx)
							break;

						case KT_DIRECTORY_INDIRECT_OFFSET:
							MyPrintf3 (VL_ALWAYS, "ind_off=%06X (%04X:%08X)", 
								dirEntry->io.indirect_offset, 
								CONFIG_ROM.HighPart, 
								CALC_LOW_PART_CFG_ROM_INDIRECT_ADDR_OFF(dirEntry->io.indirect_offset));

							// Set the referenced block info table entry to DIRECTORY type
							SET_BLOCK_INFO_INDIRECT(dirEntry->io.indirect_offset,BT_DIRECTORY,dirEntry->q,curTabIdx)
							break;
					}
				}
				// It's a leaf, just print the leaf data as characters
				else
				{
					DumpQuadletAsChars (*cr);
				}
				break;

			// We're entering a LEAF or DIRECTORY block
			case BT_LEAF:
			case BT_DIRECTORY:

				// Print the name of the leaf or directory type
				MyPrintf1 (VL_ALWAYS, "\n%s", 
					(DWORD)KeyToString(
						&bi->ReferringDirEntry,
						curUnitSpecId, 
						curUnitSwVersion));

				// Print the reference address offset (if there is one)
				// NOTE:  If we're at the root, we don't have any
				// refernce to worry about.
				if (bi->ReferringDirEntry.q != DT_ROOT)
				{
					MyPrintf2 (VL_ALWAYS, " referenced from %04X:%08X", 
						CONFIG_ROM.HighPart, 
						CONFIG_ROM.LowPart+bi->ReferringDirEntryIdx*sizeof(QUADLET));
				}
				MyPrintf (VL_ALWAYS, "\n");

				// Print current address offset and value
				PRINT_ADDR_OFF_VALUE
				PRINT_HEADER_INFO

				// Set the context for the current block (DIRECTORY or LEAF)
				// This will let the parsing loop know that all of the quadlets
				// that follow this header will be either in a DIRECTORY or LEAF.
				curBlockType = bi->BlockType;
				break;
		}
		MyPrintf (VL_ALWAYS, "\n");
	}

OK:

	// OK
	retcode = TRUE;

Exit:

	// Free resources
	if (blockInfo != NULL)
		free (blockInfo);

	DBOUT1 (LF_EXIT, "DecodeConfigRom, retcode = %08X\r\n", retcode);
	return (retcode);

InvalidSize:
	MyPrintf2 (VL_ERROR, "\n\nERROR:  Buffer size %d is not divisible by %d\n\n", CfgRomBufSize, sizeof(QUADLET));
	goto Exit;

AllocError:
	MyPrintf(VL_ERROR, "ERROR:  Cannot allocate memory.\n");
	goto Exit;

MinimalCfgRom:

	// Print minimal CFG ROM header and exit OK
	PRINT_ADDR_OFF_VALUE
	MyPrintf1 (VL_ALWAYS, "Minimal Configuration ROM, GUID = %06X\n", romHdr->mcr.vendor_id);
	retcode = TRUE;
	goto Exit;

InvalidRomHdr:
	PRINT_ADDR_OFF_VALUE
	MyPrintf (VL_ALWAYS, "Invalid Configuration ROM Header\n");
	for (INC_PLACE_HOLDERS2; cr < crEnd; INC_PLACE_HOLDERS2)
	{
		PRINT_ADDR_OFF_VALUE
		MyPrintf (VL_ALWAYS, "?????\n");
	}
	goto Exit;

	#define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DumpQuadletAsChars
//
//		Prints the specified quadlet value as an array of characters.
//
//	Entry:
//
//		q		quadlet value to print
//
//	Exit:
//		NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DumpQuadletAsChars (QUADLET q)
{
	int i;
	BYTE ba[sizeof(QUADLET)];

	// Copy the quadlet to a byte array.
	// NOTE:  If little Endian, we must swap the quadlet 
	// because we will now treat as a byte array
	HostSwapQuadlet (q, (PQUADLET)ba);

	// Print the byte array as characters
	for (i=0; i<sizeof(QUADLET); i++)
	{
		if (isprint(ba[i]))
			MyPrintf1 (VL_ALWAYS, "%c", ba[i]);
		else
			MyPrintf (VL_ALWAYS, ".");
	}
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// KeyValueToString
//
//		Translate the specified directory entry's key_value into a printable
//		string.  The translation is dependent on the specified UnitSpecId
//		and UnitSwVersion.
//
//	Entry:
//
//		DirEntry			quadlet value (directory entry) to translate
//
//		UnitSpecId		Unit_Spec_Id value of the directory entry
//
//		UnitSwVersion	Unit_Sw_Version value of the directory entry
//
//	Exit:
//
//		pointer to a character string specifying the translation of
//		the key_value in the directory entry.  This is a static string
//		and should not be modified by the calling routine.  Also, the
//		string is subject to change on the next call to this routine.
//		If you need to retain the string value, you must copy it
//		before calling this routine again.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
char *KeyToString (PDirEntryT DirEntry, DWORD UnitSpecId, DWORD UnitSwVersion)
{
	//static char bad_key_value_str[128];
	static char ret_str[128];

	// Initialize static return string
	ret_str[0] = '\0';

	// Root_Directory?
	if (DirEntry->q == DT_ROOT)
		return ("Root_Directory");

	switch (DirEntry->iv.key_value)
	{
		case Textual_Descriptor:
			strcat (ret_str, "Textual_Descriptor");
			break;
		case Bus_Dependent_Info:
			strcat (ret_str, "Bus_Dependent_Info");
			break;
		case Module_Vendor_Id:
			strcat (ret_str, "Module_Vendor_Id");
			break;
		case Module_Hw_Version:
			strcat (ret_str, "Module_Hw_Version");
			break;
		case Module_Spec_Id:
			strcat (ret_str, "Module_Spec_Id");
			break;
		case Module_Sw_Version:
			strcat (ret_str, "Module_Sw_Version");
			break;
		case Module_Dependent_Info:
			strcat (ret_str, "Module_Dependent_Info");
			break;
		case Node_Vendor_Info:
			strcat (ret_str, "Node_Vendor_Info");
			break;
		case Node_Hw_Version:
			strcat (ret_str, "Node_Hw_Version");
			break;
		case Node_Spec_Id:
			strcat (ret_str, "Node_Spec_Id");
			break;
		case Node_Sw_Version:
			strcat (ret_str, "Node_Sw_Version");
			break;
		case Node_Capabilities:
			strcat (ret_str, "Node_Capabilities");
			break;
		case Node_Unique_Id:
			strcat (ret_str, "Node_Unique_Id");
			break;
		case Node_Units_Extent:
			strcat (ret_str, "Node_Units_Extent");
			break;
		case Node_Memory_Extent:
			strcat (ret_str, "Node_Memory_Extent");
			break;
		case Node_Dependent_Info:
			strcat (ret_str, "Node_Dependent_Info");
			break;
		case Unit_Directory:
			strcat (ret_str, "Unit_Directory");
			break;
		case Unit_Spec_Id:
			strcat (ret_str, "Unit_Spec_Id");
			break;
		case Unit_Sw_Version:
			strcat (ret_str, "Unit_Sw_Version");
			break;
		case Unit_Dependent_Info:			// 0x14
		//case Logical_Unit_Number:		// 0x14 SBP-2
		//case Logical_Unit_Directory:	// 0xD4 SBP-2
		//case Management_Agent:			// 0x54 SBP-2
			if (UnitSpecId == SBP2_Unit_Spec_Id && UnitSwVersion == SBP2_Unit_Sw_Version)
			{
				switch (DirEntry->iv.key_type)
				{
					case KT_IMMEDIATE_VALUE:
						strcat (ret_str, "Logical_Unit_Number");
						break;
					case KT_DIRECTORY_INDIRECT_OFFSET:
						strcat (ret_str, "Logical_Unit_Directory");
						break;
					case KT_CSR_OFFSET:
						strcat (ret_str, "Management_Agent");
						break;
					case KT_LEAF_INDIRECT_OFFSET:
					default:
						break;

				}
			}
			else
			{
				strcat (ret_str, "Unit_Dependent_Info");
			}
			break;
		case Unit_Location:
			strcat (ret_str, "Unit_Location");
			break;
		case Unit_Poll_Mask:
			strcat (ret_str, "Unit_Poll_Mask");
			break;
		case Unit_Power_Requirements:
			strcat (ret_str, "Unit_Power_Requirements");
			break;
		case Model_Id:
			strcat (ret_str, "Model_Id");
			break;
		case Command_Set_Spec_Id:
			strcat (ret_str, "Command_Set_Spec_Id");
			break;
		case Command_Set:
			strcat (ret_str, "Command_Set");
			break;
		case Unit_Characteristics:
			strcat (ret_str, "Unit_Characteristics");
			break;
		case Command_Set_Revision:
			strcat (ret_str, "Commanmd_Set_Revision");
			break;
		case Firmware_Revision:
			strcat (ret_str, "Firmware_Revision");
			break;
//		case :
//			strcat (ret_str, "");
//			break;
		default:
			sprintf (ret_str+strlen(ret_str), "key=%02X (UNKNOWN)", DirEntry->general.key);
			break;
	}
	switch (DirEntry->iv.key_type)
	{
		case KT_LEAF_INDIRECT_OFFSET:
			strcat (ret_str, " leaf");
			break;
		case KT_DIRECTORY_INDIRECT_OFFSET:
			strcat (ret_str, " directory");
			break;
	}
	return (ret_str);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ImmediateValueToString
//
//		Translate the specified directory entry's value into a printable
//		string.  The translation is dependent on the specified UnitSpecId
//		and UnitSwVersion.
//
//	Entry:
//
//		DirEntry			quadlet value (directory entry) to translate
//
//		UnitSpecId		Unit_Spec_Id value of the directory entry
//
//		UnitSwVersion	Unit_Sw_Version value of the directory entry
//
//	Exit:
//
//		pointer to a character string specifying the translation of
//		the value in the directory entry.  This is a static string
//		and should not be modified by the calling routine.  Also, the
//		string is subject to change on the next call to this routine.
//		If you need to retain the string value, you must copy it
//		before calling this routine again.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
char *ImmediateValueToString (PDirEntryT DirEntry, DWORD UnitSpecId, DWORD UnitSwVersion)
{
	static char ret_str[128];
	DWORD tempvalue;

	// Initialize vars
	ret_str[0] = '\0';
	tempvalue = DirEntry->iv.value;

	// The meaning of the immediate value is dependent on the key_value
	switch (DirEntry->iv.key_value)
	{
		case Textual_Descriptor:
		case Bus_Dependent_Info:
		case Module_Dependent_Info:
		case Node_Unique_Id:
		case Node_Dependent_Info:
		case Unit_Directory:
		case Unit_Location:
			sprintf(ret_str, "ERROR:  No immediate value possible");
			break;
		case Module_Vendor_Id:
		case Module_Hw_Version:
		case Module_Spec_Id:
		case Module_Sw_Version:
		case Node_Vendor_Info:
		case Node_Hw_Version:
		case Node_Spec_Id:
		case Node_Sw_Version:
		case Unit_Spec_Id:
		case Unit_Sw_Version:
		case Command_Set_Spec_Id:
		case Command_Set:
		case Unit_Characteristics:
		case Command_Set_Revision:
		case Firmware_Revision:
			sprintf(ret_str, "%06X", DirEntry->iv.value);
			break;
		case Unit_Dependent_Info:		// 0x14
		//case Logical_Unit_Number:		// 0x14 SBP-2
			if (UnitSpecId == SBP2_Unit_Spec_Id && UnitSwVersion == SBP2_Unit_Sw_Version)
			{
				PLogicalUnitNumberT p;
				p = (PLogicalUnitNumberT)&tempvalue;
				sprintf(ret_str, "o=%X, device_type=%02X, lun=%04X", p->o, p->device_type, p->lun);
			}
			else
				sprintf(ret_str, "ERROR:  No immediate value possible");
			break;
		case Node_Capabilities:
			{
				PNodeCapabilitiesT p;
				p = (PNodeCapabilitiesT)&tempvalue;
				if (p->spt)
					strcat (ret_str, "spt ");
				if (p->ms)
					strcat (ret_str, "ms ");
				if (p->_int)
					strcat (ret_str, "int ");
				if (p->ext)
					strcat (ret_str, "ext ");
				if (p->bas)
					strcat (ret_str, "bas ");
				if (p->prv)
					strcat (ret_str, "prv ");
				if (p->_64)
					strcat (ret_str, "64 ");
				if (p->fix)
					strcat (ret_str, "fix ");
				if (p->lst)
					strcat (ret_str, "lst ");
				if (p->drg)
					strcat (ret_str, "drg ");
				if (p->elo)
					strcat (ret_str, "elo ");
				if (p->atn)
					strcat (ret_str, "atn ");
				if (p->off)
					strcat (ret_str, "off ");
				if (p->ded)
					strcat (ret_str, "ded ");
				if (p->init)
					strcat (ret_str, "init ");
			}
			break;
		case Node_Units_Extent:
		case Node_Memory_Extent:
			{
				PNodeUnitsExtentT p;
				p = (PNodeUnitsExtentT)&tempvalue;
				sprintf (ret_str, "align=%02X, number=%04X", p->align, p->number);
			}
			break;
		case Unit_Poll_Mask:
			{
				PUnitPollMaskT p;
				p = (PUnitPollMaskT)&tempvalue;
				sprintf (ret_str, "poll_mask=%04X", p->poll_mask);
			}
			break;
		case Unit_Power_Requirements:
			sprintf(ret_str, "%06X deciwatts", DirEntry->iv.value);
			break;
		case KV_UNKNOWN:
		default:
			sprintf(ret_str, "value = %06X", DirEntry->iv.value);
			break;
	}

	return (ret_str);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// LI_HOST_SWAP
//
//		Swaps the high and low parts of the specified LARGE_INTEGER
//
//	Entry:
//
//		li		LARGE_INTEFER to swap
//
//	Exit:
//
//		LowPart	 = li.HighPart
//		HighPart	 = li.LowPart
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
LARGE_INTEGER LI_HOST_SWAP (LARGE_INTEGER li)
{
	LARGE_INTEGER retcode;
	retcode.HighPart = li.LowPart;
	retcode.LowPart = li.HighPart;
	return (retcode);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// HostSwapQuadlet
//
//		Swaps the quadlet value according to the platform
//		we are running on.
//
//	Entry:
//
//		qValue		quadlet value to swap
//
//		dst			pointer to a quadlet variable to be filled in
//
//	Exit:
//
//		dst			swapped qaudlet value
//
//		if we are running on an Intel machine, we will swap
//		the quadlet.  Otherwise, no swap will be performed.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void HostSwapQuadlet (QUADLET qValue, PQUADLET dst)
{
	#ifdef I386

		QUADLET tmp;

		tmp = qValue;
		_asm
		{
			mov		eax, tmp;
			xchg	al, ah;
			clc;
			ror		eax, 16;
			xchg	al, ah;
			mov		tmp, eax;
		}
		*dst = tmp;

	#else

		*dst = qValue;

	#endif
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// LoadLibraryAndFixPointers 
//
//		Loads the CLS1394.DLL and fixes up the function pointers
//		to the functions we need.
//
//	Entry:
//		NONE
//
//	Exit:
//		
//		TRUE				operation succeeded
//
//		FALSE				operation failed
//
//		if TRUE, then various global function pointers
//		will be set and ready to call.
//
//	NOTE:  The reason we have to do things this way is that we
//	want to independent of 2.1 and 2.2 import link libraries.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOL LoadLibraryAndFixPointers (void)
{
	//#define DEBUGFLAGS DebugThisRoutine

	BOOL retcode;

	DBOUT (LF_ENTRY, "LoadLibraryAndFixPointers\r\n");

	// Assume failure
	retcode = FALSE;

	// Load the CLS1394.DLL
	hDll = LoadLibrary("cls1394.dll");
	if (hDll == NULL)
		goto LoadLibError;

	// Get the pointers to the functions we want
   lpfnCls1394Initialize = (PCls1394InitializeT) GetProcAddress (hDll, "cls1394Initialize");
   if (!lpfnCls1394Initialize)
		goto GetProcAddrError;
   lpfnCls1394Terminate = (PCls1394TerminateT) GetProcAddress (hDll, "cls1394Terminate");
   if (!lpfnCls1394Terminate)
		goto GetProcAddrError;
   lpfnCls1394CreateFile = (PCls1394CreateFileT) GetProcAddress (hDll, "cls1394CreateFile");
   if (!lpfnCls1394CreateFile)
		goto GetProcAddrError;
   lpfnCls1394CloseHandle = (PCls1394CloseHandleT) GetProcAddress (hDll, "cls1394CloseHandle");
   if (!lpfnCls1394CloseHandle)
		goto GetProcAddrError;
   lpfnCls1394DeviceIoControl = (PCls1394DeviceIoControlT) GetProcAddress (hDll, "cls1394DeviceIoControl");
   if (!lpfnCls1394DeviceIoControl)
		goto GetProcAddrError;
   lpfnGetDeviceInfo = (PGetDeviceInfoT) GetProcAddress (hDll, "GetDeviceInfo");
   if (!lpfnGetDeviceInfo)
		goto GetProcAddrError;
//   lpfnCls1394 = (PCls1394T) GetProcAddress (hDll, "cls1394");
//   if (!lpfnCls1394)
//		goto GetProcAddrError;

	// OK
	retcode = TRUE;

Exit:

	DBOUT1 (LF_EXIT, "LoadLibraryAndFixPointers, retcode = %08X\r\n", retcode);
	return (retcode);

LoadLibError:
	printf ("\nERROR:  Unable to load CLS1394.DLL!\n\n");
	goto Exit;

GetProcAddrError:
	printf ("\nERROR:  Unable to get the address of a procedure within CLS1394.DLL!\n\n");
	goto Exit;

	#define DEBUGFLAGS DebugFlags
}


